# This collection of functions is intended to provide an S-language level implementation of
# converters between R and Java of list objects.  By doing this at this level, we pay a
# performance penalty because we first of all have to search through the converters, and
# also the conversion is done in interpreted code and not C code.

convertListToJava =
function(x, ...)
{
  if(length(names(x)) == 0) {
        # create an array in Java and convert each element.
     val = .JavaArrayConstructor("java.lang.Object", dim = length(x))
     for(i in seq(length = length(x))) {
       .JavaSetArrayElement(val, x[[i]], i, .convert = NULL)
     }
  } else {
     val = .JNew("org.omegahat.Environment.Utils.OrderedTable", length(x))
     for(i in names(x)) {
       val$put(i, x[[i]])
     }
  }

  return(val)
}  

convertNamedListToJava <-
 function(obj, ...) {
      h <- .JNew("java.util.HashMap")
      for(n in names(obj)) {
        .Java(h, "put", n, obj[[n]])
      }
      h
  }




convertOrderedTableFromJava =
function(x, ...)
{
  ans = vector("list")
  p = x$keys()  # A java.util.Enumeration object.
      # Many calls but little memory per-call.  
  while(p$hasMoreElements()) {
    i = p$nextElement()
    ans[[i]] = x$get(i)
  }

  ans
}



convertOrderedTableFromJava =
function(x, ...)
{
  ans = vector("list")
  for(i in x$orderedKeys()) {
    ans[[i]] = x$get(i)
  }

  ans
}  


convertVectorFromJava =
function(x, ...)
{
  n = x$size()

  ans = vector("list", n)
  for(i in seq(length = n)) {
    ans[[i]] = x$elementAt(as.integer(i-1))
  }

  simplifyListToVector(ans)
}


simplifyListToVector =
  # 
  # Taken from sapply()
  #
function(answer)
{
 if(length(answer) && length(common.len <- unique(unlist(lapply(answer, length)))) == 1) {

   types = unique(sapply(answer, class))
   if(length(types) == 1 || sort(types) == c("integer", "numeric")) {
     if (common.len == 1) 
       answer = unlist(answer, recursive = FALSE)
     else if (common.len > 1) 
       answer = array(unlist(answer, recursive = FALSE), dim = c(common.len, length(answer)),
         dimnames = if (!(is.null(n1 <- names(answer[[1]])) & is.null(n2 <- names(answer)))) 
         list(n1, n2))
   }
 } 

 answer
}


convertArrayFromJava <-
function(x, ...)
{
   n = .JavaArrayLength(x)

   a = vector("list", n)

   for(i in seq(length = n)) {
        a[[i]] = x[i]
   }

   simplifyListToVector(a)
}  
