#include "ROmegahatInterpreter.h"
#include <stdlib.h>
#include <string.h>


#ifdef __cplusplus 
extern "C" {
#endif

#include "RSCommon.h"
#include <Rinternals.h>
#include <Rembedded.h>
/* extern int Rf_initEmbeddedR(int argc, char *argv[]); */
extern Rboolean R_Interactive;

void setup_Rmainloop(void);
void run_Rmainloop(void);


#include "CtoJava.h"
#include "RStoJava.h"

#include "Converters.h"

#ifdef __cplusplus 
}
#endif



#define VMENV (*env)->

/*
 The implementation of the initialization of an embedded R from Java.
*/
JNIEXPORT jboolean JNICALL Java_org_omegahat_R_Java_ROmegahatInterpreter_initR(JNIEnv *env, jobject engine, jobjectArray args)
{
  int argc = VMENV GetArrayLength(env, args);
  char **argv;
  jstring jtmp;
  const char *tmp;
  jboolean isCopy;
  int i;

   argv = (char **) malloc(sizeof(char*) * argc);

   for(i = 0 ; i < argc; i++) {
      jtmp = (jstring) VMENV GetObjectArrayElement(env, args, i);
      tmp = VMENV GetStringUTFChars(env, jtmp, &isCopy);
      argv[i] = strdup(tmp);
      if(isCopy == JNI_TRUE)
        VMENV ReleaseStringUTFChars(env, jtmp, tmp);
   }

  Rf_initEmbeddedR(argc, argv);
  /* R_tryEval will return when R_Interactive is TRUE */
  R_Interactive=TRUE;

 return(JNI_TRUE);
}



/*
 This is the simple non-terminating event loop
 for R that process the different input sources
 such as the command line and also X11 graphics device, etc.
*/
JNIEXPORT void JNICALL 
Java_org_omegahat_R_Java_ROmegahatInterpreter_run(JNIEnv *env, jobject engine)
{
 /* 
    If we call this here
       setup_Rmainloop();
     we replicate the banner message.
  */
 run_Rmainloop();
}

JNIEXPORT jboolean JNICALL 
Java_org_omegahat_R_Java_ROmegahatInterpreter_registerNativeConverters(JNIEnv *env, jobject manager)
{
  RS_JAVA(registerDefaultConverters(getJNIEnv()));
  return(JNI_TRUE);
}

#ifdef DEFINE_CALL
int eval_R_command(const char *funcName);

JNIEXPORT void JNICALL Java_RJava_Rengine_callFunc(JNIEnv *env, jobject engine, jstring func)
{
 jboolean isCopy;
 const char *funcName = VMENV GetStringUTFChars(func, &isCopy);
/* setup_Rmainloop(); */
 eval_R_command(funcName);

 run_Rmainloop();
 if(isCopy == JNI_TRUE)
     VMENV ReleaseStringUTFChars(func, funcName);

}

/*
  This is a simple version of a function call that takes
  a single argument which is constructed as `1:10'.
  It allows us to test different function calls such
  as plot(1:10), print(1:10), sum(1:10), etc.
*/
int
eval_R_command(const char *funcName)
{
 SEXP e;
 SEXP fun;
 SEXP arg;
 int i;

    fun = Rf_findFun(Rf_install((char *)funcName),  R_GlobalEnv);
    PROTECT(fun);
    PROTECT(arg = NEW_INTEGER(10));
    for(i = 0; i < GET_LENGTH(arg); i++)
      INTEGER_DATA(arg)[i]  = i + 1;

    e = allocVector(LANGSXP, 2);
    PROTECT(e);
    SETCAR(e, fun);
    SETCAR(CDR(e), arg);

      /* Evaluate the call to the R function.
         Ignore the return value.
       */
    eval(e, R_GlobalEnv);
    UNPROTECT(3);   
  return(0);
}

#endif /* DEFINE_CALL */
