/* #line 3 "REval.cweb" */
#include <Rinternals.h>
#include <R_ext/Parse.h>
#include "CtoJava.h"
#include "REvaluator.h"
#include "RtoJava.h"
#include "RStoJava.h"
#include "ErrorHandling.h"

extern void R_SetErrorHook(void (*)(SEXP, char *));

/* #line 306 "REval.cweb" */
void
JavaErrorHook(SEXP call, char *message)
{
  jclass klass;
  JNIEnv *env = getJNIEnv();

 
#if 0
  JavaVM *vm = getJavaVM();
  (*vm)->AttachCurrentThread(vm, (void**) &env, NULL);
#endif
  
  klass = VMENV FindClass(env, "org/omegahat/R/Java/RException");

#if 1  
  VMENV ThrowNew(env, klass, message);
#else
  jmethodID mid;
  jstring msg;
  jobject exception;
  
  mid = VMENV GetMethodID(env, klass, "<init>", "(Ljava/lang/String;)V");

  msg = VMENV NewStringUTF(env, message);

  exception = VMENV NewObject(env, klass, mid, msg);
  VMENV Throw(env, exception);
#endif  
}

void
sjava_RErrorHook( SEXP call, char *message )
{
  JNIEnv	*env = getJNIEnv();
  /* Record R error as (tentative -- see sjava_tryEval) Java
     exception; There might be several 'errors' if these occur in a
     try() statement that does not stop successful execution. The
     handled 'error' messages can then recovered from a Java-level
     stderr sink. TODO: There is no guarantee that existing exceptions
     are from R alone.  */
  if ( VMENV ExceptionCheck( env ) )
    VMENV ExceptionClear( env );
  throw_Java_exception( message, "org/omegahat/R/Java/RException", env );
  R_SetErrorHook( sjava_RErrorHook );
}

USER_OBJECT_
sjava_tryEval(SEXP expr, SEXP rEnv, JNIEnv *env, int *errorOccurred) {
  USER_OBJECT_	value;

  R_SetErrorHook( sjava_RErrorHook );

  PROTECT( expr );
  value = R_tryEval( expr, rEnv, errorOccurred );
  UNPROTECT(1);

  if( *errorOccurred ) {
    const char *excMsg = get_Java_exception( "R execution error", env );
    RS_lockEvaluator( env, JNI_FALSE, JNI_FALSE );
    throw_Java_exception( excMsg, "org/omegahat/R/Java/RException", env );
  } else if ( VMENV ExceptionCheck( env ) ) /* R error, but successful execution -- clear exception  */
    VMENV ExceptionClear( env );
  return( value );
}

jobject
convertResult(JNIEnv* env, SEXP val, jboolean convert, jboolean discard)
{
  SEXP anonRef;
  char *sig = NULL;
  jobject result = NULL_JAVA_OBJECT;

  if(discard == JNI_FALSE && val != NULL_USER_OBJECT) {
	if (convert == JNI_FALSE) {
#ifdef LDEBUG   
	  fprintf(stderr, "convertResult to anonymous reference\n");
	  fflush(stderr);
#endif
	  PROTECT(anonRef = createRAnonymousReference(val));
	  result =
		userReferenceToJava(env, anonRef, RS_ANONYMOUS_REFERENCE);
	  UNPROTECT(1);
	} else {
#ifdef LDEBUG    
	  fprintf(stderr, "convertResult to Java\n");
	  fflush(stderr);
#endif
	  result = s_to_java_basic(env, val, sig, &sig);
	}
  }
  return(result);
}

/* #line 18 "REval.cweb" */
JNIEXPORT jobject JNICALL
Java_org_omegahat_R_Java_REvaluator_eval(JNIEnv *env, jobject REvaluator, jstring cmd, jboolean convert, jclass returnClass, jboolean discard)
{
  /* Parse-related */
  jboolean isCopy;
  const char *str;
  USER_OBJECT_ userCmd, userExpr;
  ParseStatus status;
  char errString[20];

  /* Eval-related */
  USER_OBJECT_ userResult = NULL_USER_OBJECT;
  jobject result = NULL_JAVA_OBJECT;
  int i, n, errorOccurred = 0;

  RS_lockEvaluator(env, JNI_TRUE, JNI_FALSE);

  /* Copy cmd to userCmd */
  str = VMENV GetStringUTFChars(env, cmd, &isCopy);
  PROTECT(userCmd = NEW_CHARACTER(1));
  SET_STRING_ELT(userCmd, 0, COPY_TO_USER_STRING(str));
  if(isCopy == JNI_TRUE)
	VMENV ReleaseStringUTFChars(env, cmd, str);

  /* Parse userCmd to userExpr */
  PROTECT(userExpr =
		  R_ParseVector(userCmd, 1, &status, NULL_USER_OBJECT));
  if (PARSE_OK != status) {
	if (VMENV ExceptionCheck(env))
	  VMENV ExceptionClear(env);
	switch(status) {
	case PARSE_INCOMPLETE:
	  snprintf(errString, 20, "Parse incomplete");
	  break;
	case PARSE_ERROR:
	  snprintf(errString, 20, "Parse error");
	  break;
	case PARSE_EOF:
	  snprintf(errString, 20, "Parse end-of-file");
	  break;
	default:
	  snprintf(errString, 20, "Parse unknown error");
	  break;
	}
	throw_Java_exception(errString,
						 "org/omegahat/R/Java/RException", env);
	UNPROTECT(2);
	return(NULL_JAVA_OBJECT);
  }

  /* Evaluate userExpr, and (possibly) convert result */
  n = GET_LENGTH(userExpr);
  for(i = 0; !errorOccurred && i < n ; i++)
	userResult = sjava_tryEval(VECTOR_ELT(userExpr,i), R_GlobalEnv, env,
							   &errorOccurred);
  if (!errorOccurred) {
	PROTECT(userResult);
	result = convertResult(env, userResult, convert, discard);
	UNPROTECT(1);
	RS_lockEvaluator(env, JNI_FALSE, JNI_FALSE);
  }
  UNPROTECT(2);
  return(result);
}    

  /* #line 121 "REval.cweb" */
JNIEXPORT jobject JNICALL
Java_org_omegahat_R_Java_REvaluator_call(JNIEnv *env, jobject evaluator, jstring functionName, jobjectArray args, jobjectArray argNames, jclass targetClass, jboolean convert)
{
  int numArgs = 0;
  USER_OBJECT_ fun, e, userResult;
  jobject result = NULL_JAVA_OBJECT;
  int numProtect = 0;

  const char *str;
  jboolean isCopy;

  int errorOccurred;

#ifdef LDEBUG    
  fprintf(stderr, "In REvaluator_call:  env = %p\n", env); fflush(stderr);
  if(args == NULL_JAVA_OBJECT) {
	fprintf(stderr, "No arguments\n"); fflush(stderr);
  }
#endif 

  if(args != NULL_JAVA_OBJECT)
	numArgs = VMENV GetArrayLength(env, args);

  RS_lockEvaluator(env, JNI_TRUE, JNI_FALSE); 


#ifdef LDEBUG    
  fprintf(stderr, "Expression length %d\n", numArgs); fflush(stderr);
#endif 
  PROTECT(e = allocVector(LANGSXP, numArgs + 1));
  numProtect++;

 
  str = VMENV GetStringUTFChars(env, functionName, &isCopy);

#if 0
  /* We would like to be able to do this so that we can get more refined error messages.
     However, we cannot easily control the error handling at this point.
     So we try to eval the entire expression eventhough we could bail here if we
     knew the function wasn't there.
  */
  fun = Rf_findFun(Rf_install((char *)str),  R_GlobalEnv);
#else 
  fun = Rf_install(str);
#endif
  if(isCopy == JNI_TRUE) {
	VMENV ReleaseStringUTFChars(env, functionName, str);
  }  
  SETCAR(e, fun);

  if(numArgs > 0) {
	/* Add the ordered and named arguments to the expression. */
    int i;
    const char *keyName = NULL;
    jobject argVal, argName;
    USER_OBJECT_ tmp, cvt;
#ifdef LDEBUG       
	fprintf(stderr, "Number of arguments %d\n", numArgs);fflush(stderr);
#endif
    
    tmp = CDR(e);
    for(i = 0; i < numArgs; i++) {

#ifdef LDEBUG           
	  fprintf(stderr, " processing arg %d)\n", i);fflush(stderr);
#endif
      argVal = VMENV GetObjectArrayElement(env, args, i);
      
#ifdef LDEBUG   
	  fprintf(stderr, " converting arg %d %p\n", i, argVal);fflush(stderr);
#endif
	  cvt = s_from_java_basic(env, argVal, NULL, JNI_FALSE, JNI_TRUE);
	  SETCAR(tmp, cvt);     
      
      argName= VMENV GetObjectArrayElement(env, argNames, i);
      if(argName == NULL_JAVA_OBJECT)
		keyName = NULL;
      else {
#ifdef LDEBUG                     
		fprintf(stderr, "getting argument name %d) %p\n", i, argName);fflush(stderr);
#endif   
		keyName = VMENV GetStringUTFChars(env, argName, &isCopy);
#ifdef LDEBUG                    
		fprintf(stderr, "argument name %d) %s\n", i, keyName);fflush(stderr);
#endif   
      }


#ifdef LDEBUG   
	  fprintf(stderr, " converted arg %d\n", i);fflush(stderr);
#endif

      if(keyName && keyName[0]) {
        SET_TAG(tmp, Rf_install(keyName));
        if(isCopy == JNI_TRUE) {
          VMENV ReleaseStringUTFChars(env, argName, keyName);
        }
      }
      
      tmp = CDR(tmp);
    }
  }

#ifdef LDEBUG   
  fprintf(stderr, "evaluating function call\n"); fflush(stderr);
#endif 

  /* Evaluate the call and (possibly) convert the result */
  userResult = sjava_tryEval(e, R_GlobalEnv, env, &errorOccurred);
  if(!errorOccurred) {
	PROTECT(userResult);numProtect++;
	result = convertResult(env, userResult, convert, JNI_FALSE);
	RS_lockEvaluator(env, JNI_FALSE, JNI_FALSE); 
  }

  /* Tidy and return */
  UNPROTECT(numProtect);
#ifdef LDEBUG   
  fprintf(stderr, "completed native REvaluator.call()\n");fflush(stderr);
#endif 
  return(result); 
}


/* #line 270 "REval.cweb" */
jboolean
JNICALL 
Java_org_omegahat_Interfaces_NativeInterface_OmegaInterfaceManager_nativeSetup(
              JNIEnv *env, jclass Class, jobject This, jobject nullObj) 
{
extern void RS_JAVA_registerDefaultConverters(JNIEnv *env);
extern int R_CStackLimit;
  setJNIEnv(env);
  computeJavaVM(env);
  do_native_setup(This, nullObj, env);

  R_SetErrorHook(JavaErrorHook);
  R_CStackLimit = -1;

/*
   This is done in the Java code so no need to do it here.
   RS_JAVA_registerDefaultConverters(env);
*/


  return(JNI_TRUE);
}
